import networkx as nx

from paraffin.db.app import export_db_to_graph, save_graph_to_db
from paraffin.dvc import get_status


def test_db_graph_conversion(proj01):
    db_path = "sqlite:///paraffin.db"
    status_graph: nx.DiGraph = get_status()

    save_graph_to_db(
        db_url=db_path,
        graph=status_graph,
    )

    db_graph: nx.DiGraph = export_db_to_graph(db_url=db_path)

    # You can directly compare nodes and edges
    assert len(status_graph) == 14
    assert len(status_graph.edges) == 12

    assert set(status_graph.nodes) == set(db_graph.nodes)
    assert set(status_graph.edges) == set(db_graph.edges)



import os
import socket
import threading
import typing as t

import typer

from paraffin.db.app import (
    close_worker,
    get_job,
    register_worker,
    update_job,
)

app = typer.Typer()


@app.command()
def commit():
    """Commit all reproduced stages."""
    import json

    import dvc.api
    from dvc.stage.serialize import to_single_stage_lockfile

    from paraffin.dvc import StageStatus

    name = "paraffin"
    db = "sqlite:///paraffin.db"

    worker_id = register_worker(
        name=name,
        machine=socket.gethostname(),
        db_url=db,
        cwd=os.getcwd(),
        pid=os.getpid(),
        requires_dvc_lock=True,
    )
    # TODO: make this a DVC worker and ensure only one worker is running at a time
    active_job = None
    fs = dvc.api.DVCFileSystem()
    while True:
        try:
            res = get_job(
                db_url=db,
                queues=None,
                worker_id=worker_id,
                experiment=None,
                stage_name=None,
                status=[StageStatus.FINISHED],
            )
            if res is None:
                break

            stage, job = res
            active_job = job
            pipelinestage = list(
                fs.repo.stage.collect(stage.name)
            )  # TODO: does this work with path?
            if not pipelinestage:
                raise ValueError(f"Stage '{stage.name}' not found in DVC pipeline.")

            with pipelinestage[0].repo.lock:
                pipelinestage[0].save()
                pipelinestage[0].commit()
                pipelinestage[0].dump(update_pipeline=True, update_lock=True)

            update_job(
                db_url=db,
                stage_id=job.stage_id,
                status=StageStatus.COMPLETED,
                lockfile=json.dumps(
                    to_single_stage_lockfile(pipelinestage[0], with_files=True)
                ),
            )
            active_job = None
        finally:
            if active_job:
                update_job(
                    db_url=db,
                    stage_id=active_job.stage_id,
                    status=StageStatus.FINISHED,
                )
                active_job = None

    print("No job found.")
    close_worker(id=worker_id, db_url=db)


@app.command()
def worker(
    queues: str = typer.Option(
        "default",
        "--queues",
        "-q",
        envvar="PARAFFIN_QUEUES",
        help="Comma separated list of queues to listen on.",
    ),
    name: str = typer.Option(
        "default", "--name", "-n", help="Specify a custom name for this worker."
    ),
    stage: str | None = typer.Option(None, help="Job ID to run."),
    experiment: str | None = typer.Option(
        None, "--experiment", "-e", help="Experiment ID to run."
    ),
    timeout: int = typer.Option(
        0,
        "--timeout",
        "-t",
        help="Timeout in seconds before exiting"
        " the worker if no more jobs are in the queue.",
    ),
    db: str = typer.Option(
        "sqlite:///paraffin.db", help="Database URL.", envvar="PARAFFIN_DB"
    ),
    jobs: int = typer.Option(1, "--jobs", "-j", help="Number of jobs to run."),
):
    """Start a paraffin worker to process the queued DVC stages."""
    import signal

    from paraffin.worker import run_worker

    shutdown_event = threading.Event()

    def handle_shutdown(*args, **kwargs):
        shutdown_event.set()

    signal.signal(signal.SIGINT, handle_shutdown)
    signal.signal(signal.SIGTERM, handle_shutdown)

    threads = []
    for _ in range(jobs):
        t = threading.Thread(
            target=run_worker,
            args=(name, db, shutdown_event),
            daemon=True,
        )
        threads.append(t)
        t.start()

    for t in threads:
        t.join()
    print("All workers done.")


@app.command()
def submit(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output."),
    cache: bool = typer.Option(
        False,
        help="Use the paraffin cache in addition to the DVC cache"
        " to checkout cached jobs.",
    ),
    db: str = typer.Option(
        "sqlite:///paraffin.db", help="Database URL.", envvar="PARAFFIN_DB"
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="reproduce pipelines, regenerating its results, even if no changes"
        " were found. See https://dvc.org/doc/command-reference/repro#-f"
        " for more information.",
    ),
    single_item: bool = typer.Option(
        False,
        "--single-item",
        "-s",
        help="reproduce only a single stage by turning off the recursive search for"
        " changed dependencies. See https://dvc.org/doc/command-reference/repro#-s"
        " for more information.",
    ),
    # TODO: cleanup
):
    """Run DVC stages in parallel."""
    # imports here for better performance
    from paraffin.db.app import save_graph_to_db, update_existing_experiment_stages
    from paraffin.dvc import cleanup_stages, get_status, print_graph_description
    from paraffin.utils import handle_existing_stages

    # TODO: if there is an experiment, set the stages to outdated

    graph = get_status()
    handle_existing_stages(graph=graph, db=db)
    update_existing_experiment_stages(db_url=db)
    cleanup_stages(graph=graph)
    # cleanup all stages that are `queued`
    print_graph_description(graph)
    save_graph_to_db(graph=graph, db_url=db)


@app.command()
def status():
    from paraffin.dvc import get_status, print_graph_description
    # TODO: status will also perform checkouts!!

    graph = get_status()
    print_graph_description(graph)



from pathlib import Path

import zntrack
from typer.testing import CliRunner

from paraffin.cli import app

runner = CliRunner()


class CleanupCheckNode(zntrack.Node):
    text: str = zntrack.params()
    file: Path = zntrack.outs_path(zntrack.nwd / "output.txt")
    metrics_file: Path = zntrack.metrics_path(zntrack.nwd / "metrics.json")
    plots_file: Path = zntrack.plots_path(zntrack.nwd / "plots.json")

    def run(self):
        assert not self.file.exists()
        assert not self.metrics_file.exists()
        assert not self.plots_file.exists()

        self.file.parent.mkdir(parents=True, exist_ok=True)
        self.file.write_text(self.text)
        self.metrics_file.write_text(self.text)
        self.plots_file.write_text(self.text)


def test_run_cleanup(proj_path, check_finished):
    project = zntrack.Project()

    with project:
        node = CleanupCheckNode(text="Hello World!")

    project.build()

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["commit"])
    assert result.exit_code == 0
    assert check_finished()

    node.text = "Lorem Ipsum"
    project.build()
    assert node.file.exists()

    # Now we are testing the cleanup
    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    assert not node.file.exists()  # < file should be deleted at submit

    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0
    result = runner.invoke(app, ["commit"])
    assert result.exit_code == 0
    assert check_finished()



# import datetime
# import fnmatch
import datetime
import typing as t

import networkx as nx

# import networkx as nx
# from dvc.stage.cache import _get_cache_hash
from sqlmodel import (
    Session,
    SQLModel,
    create_engine,
    select,
    text,
)

from paraffin.db.models import (
    Experiment,
    ExperimentStatus,
    Job,
    Stage,
    StageDependency,
    Worker,
    WorkerStatus,
)
from paraffin.dvc import StageDC, StageStatus

# from paraffin.lock import clean_lock
# from paraffin.stage import PipelineStageDC
# from paraffin.utils import get_group


def query_existing_experiments(
    db_url: str, status: StageStatus, graph: nx.DiGraph
) -> list[Stage]:
    # TODO
    commit = "test"
    origin = "test"
    machine = "test"

    engine = create_engine(db_url)
    SQLModel.metadata.create_all(engine)

    stages = []

    with Session(engine) as session:
        statement = select(Experiment).where(
            Experiment.base == commit,
            Experiment.origin == origin,
            Experiment.machine == machine,
            Experiment.status == ExperimentStatus.ACTIVE,
        )
        results = session.exec(statement)
        experiments = results.all()
        for experiment in experiments:
            # find all jobs that are running, unfinished or finished
            statement = select(Stage).where(
                Stage.experiment_id == experiment.id,
                Stage.status == status,
                Stage.name.in_([node.addressing for node in graph]),
            )
            results = session.exec(statement)
            stages.extend(results.all())
    return stages


def update_existing_experiment_stages(db_url: str) -> None:
    # TODO: instead of updating the stages we can keep that information and update the experiment!
    commit = "test"
    origin = "test"
    machine = "test"

    engine = create_engine(db_url)
    SQLModel.metadata.create_all(engine)

    with Session(engine) as session:
        statement = select(Experiment).where(
            Experiment.base == commit,
            Experiment.origin == origin,
            Experiment.machine == machine,
            Experiment.status == ExperimentStatus.ACTIVE,
        )
        results = session.exec(statement).all()
        for experiment in results:
            # find all jobs that are running, unfinished or finished
            experiment.status = ExperimentStatus.INACTIVE
            session.add(experiment)
        session.commit()


def save_graph_to_db(graph: nx.DiGraph, db_url: str) -> None:
    engine = create_engine(db_url)
    SQLModel.metadata.create_all(engine)

    # TODO
    commit = "test"
    origin = "test"
    machine = "test"

    with Session(engine) as session:
        experiment = Experiment(base=commit, origin=origin, machine=machine)
        session.add(experiment)
        session.commit()

        for node in nx.topological_sort(graph):
            node: StageDC
            if node.cmd is None:
                continue  # skip everything that is not a PipelineStage
            job = Stage(
                cmd=node.cmd,
                name=node.addressing,
                queue="default",
                status=node.status,
                experiment_id=experiment.id,
                cache=False,
                force=False,
                path=node.path,
                lockfile_content=node.lockfile,
            )
            session.add(job)

            for parent in graph.predecessors(node):
                parent_job = session.exec(
                    select(Stage)
                    .where(Stage.experiment_id == experiment.id)
                    .where(Stage.name == parent.addressing)
                ).all()
                if len(parent_job) == 1:
                    # if the previous stage is not PipelineStage, we skip it
                    session.add(
                        StageDependency(parent_id=parent_job[0].id, child_id=job.id)
                    )
        session.commit()


def update_job(
    db_url: str,
    stage_id: int,
    status: StageStatus,
    **kwargs: t.Any,
) -> None:
    """
    Update the status of a job in the database.
    """
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = select(Stage).where(Stage.id == stage_id)
        results = session.exec(statement)
        stage = results.one()
        stage.status = status
        for key, value in kwargs.items():
            if hasattr(stage, key):
                setattr(stage, key, value)
        session.add(stage)
        session.commit()


def claim_stage(session: Session, status: list[StageStatus]) -> t.Optional[Stage]:
    # TODO
    commit = "test"
    origin = "test"
    machine = "test"

    result = session.exec(
        text(f"""
        UPDATE stage
        SET status = '{StageStatus.RUNNING}'
        WHERE id = (
            SELECT s.id FROM stage s
            JOIN experiment e ON s.experiment_id = e.id
            WHERE s.status IN ({",".join(f"'{s}'" for s in status)})
              AND e.status = '{ExperimentStatus.ACTIVE}'
              AND e.base = '{commit}'
              AND e.machine = '{machine}'
              AND e.origin = '{origin}'
            LIMIT 1
        )
        RETURNING id
    """),
    )
    row = result.first()
    if row:
        return session.exec(select(Stage).where(Stage.id == row[0])).one()
    return None


def get_job(
    db_url: str,
    worker_id: int,
    status: list[StageStatus],
    queues: list | None = None,
    experiment: int | None = None,
    stage_name: str | None = None,
) -> tuple[Stage, Job] | None:
    engine = create_engine(db_url)
    with Session(bind=engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == worker_id)).one()
        stage = claim_stage(session, status=status)

        if stage and _all_parents_completed(stage):
            job = stage.attach_job(worker)
            session.add(job)
            session.add(stage)
            session.commit()
            session.refresh(stage)
            session.refresh(job)
            return stage, job

    return None


def _all_parents_completed(stage: Stage) -> bool:
    """
    Check if all parents of a job are completed.
    """
    return all(
        parent.status in [StageStatus.COMPLETED, StageStatus.FINISHED]
        for parent in stage.parents
    )


def register_worker(
    name: str,
    machine: str,
    db_url: str,
    cwd: str,
    pid: int,
    requires_dvc_lock: bool = False,
) -> int:
    engine = create_engine(db_url)
    with Session(engine) as session:
        worker = Worker(
            name=name,
            machine=machine,
            cwd=cwd,
            pid=pid,
            requires_dvc_lock=requires_dvc_lock,
        )
        session.add(worker)
        session.commit()
        return worker.id


def close_worker(id: int, db_url: str) -> None:
    engine = create_engine(db_url)
    with Session(engine) as session:
        worker = session.exec(select(Worker).where(Worker.id == id)).one()
        worker.status = WorkerStatus.OFFLINE
        worker.last_seen = datetime.datetime.now()
        worker.finished_at = datetime.datetime.now()
        session.add(worker)
        session.commit()


def get_stage_status(
    db_url: str,
    stage_name: str | None = None,
) -> StageStatus:
    """
    Get the status of a stage in the database.
    """
    engine = create_engine(db_url)
    with Session(engine) as session:
        statement = select(Stage).where(Stage.name == stage_name)
        results = session.exec(statement)
        stage = results.one()
        return stage.status


def export_db_to_graph(db_url: str, experiment_id: int = 1) -> nx.DiGraph:
    engine = create_engine(db_url)
    SQLModel.metadata.create_all(engine)

    with Session(engine) as session:
        statement = select(Stage).where(Stage.experiment_id == experiment_id)
        results = session.exec(statement)
        stages = results.all()

        graph = nx.DiGraph()
        stage_nodes = {}

        # Create and store StageDC nodes
        for stage in stages:
            node = StageDC(
                addressing=stage.name,
                status=stage.status,
                cmd=stage.cmd,
                path=stage.path,
                lockfile=stage.lockfile_content,
            )
            graph.add_node(node)
            stage_nodes[stage.name] = node

        # Add edges between StageDC nodes
        for stage in stages:
            for parent in stage.parents:
                graph.add_edge(stage_nodes[parent.name], stage_nodes[stage.name])

    return graph


def list_experiments(db_url: str) -> list[dict]:
    # return [{"created_at": 1234567890, "base": "test", "origin": "test", "id": "1", "machine": "test"}]
    engine = create_engine(db_url)

    with Session(engine) as session:
        statement = select(Experiment)
        results = session.exec(statement)
        experiments = results.all()
        return [
            {
                "created_at": experiment.created_at,
                "base": experiment.base,
                "origin": experiment.origin,
                "id": experiment.id,
                "machine": experiment.machine,
                "status": experiment.status,
            }
            for experiment in experiments
        ]


def list_stages(db_url: str, experiment_id: int) -> list[dict]:
    engine = create_engine(db_url)

    with Session(engine) as session:
        statement = select(Stage).where(Stage.experiment_id == experiment_id)
        results = session.exec(statement)
        stages = results.all()
        return [
            {
                "id": stage.id,
                "name": stage.name,
                "status": stage.status,
            }
            for stage in stages
        ]


def get_stage_by_id(db_url: str, stage_id: int) -> StageDC:
    engine = create_engine(db_url)

    with Session(engine) as session:
        statement = select(Stage).where(Stage.id == stage_id)
        results = session.exec(statement)
        stage = results.one()
        return StageDC(
            addressing=stage.name,
            status=stage.status,
            cmd=stage.cmd,
            path=stage.path,
            lockfile=stage.lockfile_content,
        )



import json
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path

import dvc.api
import networkx as nx
from dvc.stage import PipelineStage
from dvc.stage.cache import RunCacheNotFoundError
from dvc.stage.serialize import to_single_stage_lockfile
from tqdm import tqdm


class StageStatus(StrEnum):
    """Stage status enum.

    Attributes
    ----------
    PENDING : str
        The stage is in the dvc.yaml but has not been run yet.
    COMPLETED : str
        The stage has been run and the output files are up to date.
        The stage is cached and the dvc.lock file is up to date.
    RUNNING : str
        The stage is currently running on a worker.
    UNFINISHED : str
        The stage is not running but has not been finished yet.
        A worker should pick it up and continue running it
        from the last checkpoint.
    FAILED : str
        The stage has failed and will not be run again.
    FINISHED : str
        The stage has been reproduced and the output files are up to date.
        The stage is not yet cached and the dvc.lock file is not up to date.
    UNKNOWN : str
        The stage has not been run yet.
        One or more dependencies have not been run yet as well.
        Therefore, the state can not be determined, because if all
        dependencies yield cached outputs, the stage might be
        in the run cache.
        Currently, this is the same as PENDING and hashed dependency
        will not be accounted for.
    """

    PENDING = "pending"
    COMPLETED = "completed"
    FINISHED = "finished"
    RUNNING = "running"
    UNFINISHED = "unfinished"
    FAILED = "failed"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class StageDC:
    addressing: str
    status: StageStatus
    cmd: str | None
    path: str
    lockfile: str | None


def get_stage_from_graph(graph: nx.DiGraph, stage: str) -> StageDC:
    for node in graph.nodes:
        if node.addressing == stage:
            return node
    raise ValueError(f"Stage {stage} not found in graph")


def _create_stage_dc(stage, status: StageStatus) -> StageDC:
    is_pipeline = isinstance(stage, PipelineStage)
    return StageDC(
        addressing=stage.addressing,
        status=status,
        cmd=json.dumps(stage.cmd) if is_pipeline else None,
        path=Path(stage.path_in_repo).parent.as_posix(),
        lockfile=json.dumps(to_single_stage_lockfile(stage, with_files=True))
        if is_pipeline
        else None,
    )


def _restore_and_classify(stage, run_cache: bool) -> StageDC:
    if run_cache:
        try:
            with stage.repo.lock:
                stage.repo.stage_cache.restore(stage, dry=False)
                stage.save()
                stage.dump()
                return _create_stage_dc(stage, StageStatus.COMPLETED)
        except (RunCacheNotFoundError, FileNotFoundError):
            return _create_stage_dc(stage, StageStatus.PENDING)
    else:
        return _create_stage_dc(stage, StageStatus.PENDING)


def get_status(run_cache: bool = True, **kwargs) -> nx.DiGraph:
    fs = dvc.api.DVCFileSystem(**kwargs)
    repo = fs.repo
    graph = repo.index.graph.reverse(copy=True)
    status = repo.status()

    results = {}
    for stage in tqdm(
        nx.topological_sort(graph),
        total=len(graph),
        desc="Checking stage status",
        unit="stage",
    ):
        if stage.addressing in status:
            results[stage] = _restore_and_classify(stage, run_cache)
        else:
            # Handle stages not in DVC status
            deps = list(graph.predecessors(stage))
            # convert to for loop to avoid long convoluted list comprehension
            if any(
                results.get(
                    dep, StageDC("", StageStatus.UNKNOWN, None, "", None)
                ).status
                != StageStatus.COMPLETED
                for dep in deps
            ):
                results[stage] = _create_stage_dc(stage, StageStatus.UNKNOWN)
            else:
                results[stage] = _create_stage_dc(stage, StageStatus.COMPLETED)

    assert len(results) == len(graph), (
        f"Expected {len(graph)} results, got {len(results)}"
    )
    return nx.relabel_nodes(graph, results, copy=True)


def print_graph_description(graph: nx.DiGraph):
    # TODO: read from database and not from dvc graph - this way the command can also be watdched
    from rich import box
    from rich.console import Console
    from rich.table import Table

    console = Console()
    table = Table(title="DVC Pipeline Stage Status", box=box.SIMPLE_HEAVY)
    table.add_column("Stage", justify="left", style="cyan", no_wrap=True)
    table.add_column("Status", justify="center", style="bold")

    status_icons = {
        StageStatus.COMPLETED: "[green]✅ Finished[/green]",
        StageStatus.FINISHED: "[green]✅ Finished[/green]",
        StageStatus.PENDING: "[yellow]🕐 Pending[/yellow]",
        StageStatus.RUNNING: "[blue]🔄 Running[/blue]",
        StageStatus.UNFINISHED: "[orange1]⏳ Unfinished[/orange1]",
        StageStatus.FAILED: "[red]❌ Failed[/red]",
        StageStatus.UNKNOWN: "[red]❓ Unknown[/red]",
    }

    for stage in graph.nodes:
        desc = status_icons.get(stage.status, f"[red]❓ Unknown ({stage.status})[/red]")
        table.add_row(stage.addressing, desc)

    console.print(table)


def cleanup_stages(graph: nx.DiGraph) -> None:
    import dvc.api

    fs = dvc.api.DVCFileSystem()
    # collect all stages that are queued
    stage_addressings = [
        stage.addressing
        for stage in graph.nodes
        if stage.status in [StageStatus.PENDING, StageStatus.UNKNOWN]
    ]

    stages = sum(
        (fs.repo.stage.collect(with_deps=False, target=ad) for ad in stage_addressings),
        [],
    )
    assert len(stages) == len(stage_addressings)
    for stage in tqdm(
        stages,
        total=len(stages),
        desc="Cleaning up stages",
        unit="stage",
    ):
        with stage.repo.lock:
            stage.remove_outs()



import zntrack
from typer.testing import CliRunner

from paraffin.cli import app
from paraffin.db.app import get_stage_status
from paraffin.dvc import StageStatus

runner = CliRunner()


class FailingNode(zntrack.Node):
    def run(self):
        raise Exception("This is a test exception")


def test_run_fails(proj_path):
    project = zntrack.Project()

    with project:
        failing_node = FailingNode()

    project.build()

    result = runner.invoke(app, "submit")
    assert result.exit_code == 0
    result = runner.invoke(app, ["worker"])
    assert result.exit_code == 0

    status = get_stage_status(
        db_url="sqlite:///paraffin.db", stage_name=failing_node.name
    )
    assert status == StageStatus.FAILED



import json
import os
import socket
import subprocess
import threading
import time
from typing import Optional

from paraffin.db.app import (
    Job,
    StageStatus,
    close_worker,
    get_job,
    register_worker,
    update_job,
)


def run_worker(name: str, db: str, shutdown_event: threading.Event):
    active_job: Optional[Job] = None

    worker_id = register_worker(
        name=name,
        machine=socket.gethostname(),
        db_url=db,
        cwd=os.getcwd(),
        pid=os.getpid(),
    )

    try:
        while not shutdown_event.is_set():
            res = get_job(
                db_url=db,
                queues=None,
                worker_id=worker_id,
                experiment=None,
                stage_name=None,
                status=[StageStatus.PENDING, StageStatus.UNKNOWN],
            )
            if res is None:
                break

            stage, job = res
            active_job = job

            cmd = json.loads(stage.cmd)
            print(f"({worker_id}) Running command: {cmd}")
            try:
                # subprocess.check_call(cmd, shell=True)
                proc = subprocess.Popen(
                    cmd,
                    shell=True,
                    preexec_fn=os.setsid,
                    universal_newlines=True,
                    cwd=stage.path,
                )
                # Wait for the process to finish but also check for shutdown
                while proc.poll() is None and not shutdown_event.is_set():
                    time.sleep(0.1)
                # If the shutdown event is set, terminate the process
                if shutdown_event.is_set():
                    proc.terminate()
                    proc.wait()
                    break
                # Check the return code
                if proc.returncode != 0:
                    raise subprocess.CalledProcessError(proc.returncode, cmd)
                update_job(
                    db_url=db,
                    stage_id=job.stage_id,
                    status=StageStatus.FINISHED,
                )
            except subprocess.CalledProcessError:
                print(f"({worker_id}) Command failed: {cmd}")
                update_job(
                    db_url=db,
                    stage_id=job.stage_id,
                    status=StageStatus.FAILED,
                )
            active_job = None
    finally:
        if active_job is not None:
            print(f"({worker_id}) Job {active_job.id} was interrupted.")
            update_job(
                db_url=db,
                stage_id=active_job.stage_id,
                status=StageStatus.UNFINISHED,
            )
        close_worker(id=worker_id, db_url=db)
        print(f"({worker_id}) Worker closed.")



import os
from pathlib import Path

from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles

from paraffin.db.app import get_stage_by_id, list_experiments, list_stages

# from paraffin.db import (
#     db_to_graph,
#     get_job_dump,
#     get_jobs,
#     list_experiments,
#     list_workers,
#     update_job_status,
# )
# from paraffin.utils import build_elk_hierarchy

FILE = Path(__file__)

app = FastAPI()

app.mount("/ui", StaticFiles(directory=FILE.parent.parent / "static"), name="ui")
app.mount(
    "/assets",
    StaticFiles(directory=FILE.parent.parent / "static" / "assets"),
    name="assets",
)


@app.get("/")
def read_root():
    return RedirectResponse(url="/ui/index.html")


@app.get("/api/v1/experiments")
def read_experiments():
    db_url = os.environ["PARAFFIN_DB"]
    return list_experiments(db_url=db_url)


@app.get("/api/v1/stages")
def read_stages(experiment: str):
    db_url = os.environ["PARAFFIN_DB"]
    return list_stages(experiment_id=int(experiment), db_url=db_url)


@app.get("/api/v1/stage")
def read_stage(id: int):
    db_url = os.environ["PARAFFIN_DB"]
    return get_stage_by_id(stage_id=id, db_url=db_url)


# @app.get("/api/v1/graph")
# def read_graph(experiment: str):
#     db_url = os.environ["PARAFFIN_DB"]

#     g = db_to_graph(experiment_id=int(experiment), db_url=db_url)
#     return build_elk_hierarchy(g)


# @app.get("/api/v1/spawn")
# def spawn(
#     name: str | None = None, experiment: int | None = None, stage: str | None = None
# ):
#     # Build the command
#     cmd = ["paraffin", "worker"]
#     if name is not None:
#         cmd += ["--name", name]
#     if experiment is not None:
#         cmd += ["--experiment", str(experiment)]
#     if stage is not None:
#         cmd += ["--stage", stage]

#     # open subprocess and forget about it
#     subprocess.Popen(cmd)
#     return 0

#     # works but is slow
#     # print(f"Running command: {' '.join(cmd)}")

#     # # Start the subprocess and capture stdout and stderr
#     # process = subprocess.Popen(
#     #     cmd,
#     #     stdout=subprocess.PIPE,
#     #     stderr=subprocess.PIPE,
#     #     text=True,  # Ensure output is decoded as text
#     #     bufsize=1,  # Line-buffered output
#     #     universal_newlines=True,
#     # )

#     # # Define a generator to stream the output
#     # def generate():
#     #     while True:
#     #         # Read stdout line by line
#     #         output = process.stdout.readline()
#     #         if output:
#     #             yield f"stdout: {output}"
#     #         else:
#     #             break

#     #         # Read stderr line by line
#     #         error = process.stderr.readline()
#     #         if error:
#     #             yield f"stderr: {error}"
#     #         else:
#     #             break

#     #     # Wait for the process to complete
#     #     process.wait()
#     #     yield f"Process completed with return code: {process.returncode}"

#     # # Return a StreamingResponse to stream the output
#     # return StreamingResponse(generate(), media_type="text/plain")


# @app.get("/api/v1/job")
# def read_job(name: str, experiment: int):
#     db_url = os.environ["PARAFFIN_DB"]
#     return get_job_dump(job_name=name, experiment_id=int(experiment), db_url=db_url)


# # list finished, running, and pending and failed jobs
# @app.get("/api/v1/jobs")
# def read_jobs(experiment: int):
#     db_url = os.environ["PARAFFIN_DB"]
#     return get_jobs(experiment_id=int(experiment), db_url=db_url)


# @app.get("/api/v1/workers")
# def read_workers(id: int | None = None):
#     db_url = os.environ["PARAFFIN_DB"]
#     return list_workers(db_url=db_url, id=id)


# # update_job_status
# @app.get("/api/v1/job/update")
# def update_job(name: str, experiment: int, status: str, force: bool = False):
#     db_url = os.environ["PARAFFIN_DB"]
#     print(
#         f"Updating job {name} to {status} in experiment {experiment} with force={force}"
#     )
#     return update_job_status(
#         job_name=name,
#         experiment_id=int(experiment),
#         status=status,
#         db_url=db_url,
#         force=force,
#     )



from datetime import datetime
from enum import StrEnum
from typing import List, Optional

from sqlmodel import Field, Relationship, SQLModel, String, UniqueConstraint

from paraffin.dvc import StageStatus


class ExperimentStatus(StrEnum):  # TODO: could be a bool
    ACTIVE = "active"
    INACTIVE = "inactive"


class WorkerStatus(StrEnum):
    """Worker status enum.

    Attributes
    ----------
    RUNNING : str
        The worker is currently running a job.
    IDLE : str
        The worker is idle and waiting for a job.
    OFFLINE : str
        The worker is offline and not available for jobs.
    """

    RUNNING = "running"
    IDLE = "idle"
    OFFLINE = "offline"


class Worker(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str = Field(max_length=100, index=True)
    machine: str = Field(max_length=100)
    status: WorkerStatus = Field(sa_type=String, default=WorkerStatus.IDLE, index=True)
    last_seen: datetime = Field(default_factory=datetime.now)
    cwd: str = Field(default="", max_length=255)  # Current working directory
    pid: int = Field(default=0)  # Process ID
    started_at: datetime = Field(default_factory=datetime.now)
    finished_at: Optional[datetime] = None
    requires_dvc_lock: bool = Field(default=False)

    # Relationships
    jobs: List["Job"] = Relationship(back_populates="worker")


class StageDependency(SQLModel, table=True):
    parent_id: int = Field(foreign_key="stage.id", primary_key=True)
    child_id: int = Field(foreign_key="stage.id", primary_key=True)

    # Unique constraint to prevent duplicate dependencies
    __table_args__ = (
        UniqueConstraint("parent_id", "child_id", name="unique_dependency"),
    )


class Experiment(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    base: str = Field()
    origin: str = Field(default="local")
    machine: str = Field(default="local")
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)
    status: ExperimentStatus = Field(
        sa_type=String, default=ExperimentStatus.ACTIVE
    )  # Status of the experiment
    # Relationships
    stages: List["Stage"] = Relationship(back_populates="experiment")


class Job(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    stage_id: int = Field(foreign_key="stage.id")
    worker_id: int = Field(foreign_key="worker.id")
    stderr: str = Field(default="")
    stdout: str = Field(default="")
    started_at: datetime = Field(default_factory=datetime.now)
    finished_at: Optional[datetime] = None

    # Relationships
    stage: Optional["Stage"] = Relationship(back_populates="jobs")
    worker: Optional[Worker] = Relationship(back_populates="jobs")


class Stage(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str = Field(max_length=100)
    cmd: str = Field(max_length=255)  # Command to execute
    status: StageStatus = Field(sa_type=String, default=StageStatus.PENDING)
    queue: str = Field(default="default", max_length=100)
    lockfile_content: str = Field(default="")  # JSON string of lockfile
    dependency_hash: str = Field(default="")  # Hash of the dependencies
    experiment_id: int = Field(foreign_key="experiment.id")
    capture_stderr: bool = Field(default=True)
    capture_stdout: bool = Field(default=True)
    started_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    cache: bool = Field(default=False)  # Use the paraffin cache for this job
    force: bool = Field(default=False)  # Rerun the job even if cached
    max_workers: int = Field(default=1)  # Maximum number of workers for this job
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)
    path: str = Field(default=".")  # Path to the dvc.yaml file

    # Relationships
    experiment: Optional[Experiment] = Relationship(back_populates="stages")
    jobs: List[Job] = Relationship(back_populates="stage")
    parents: List["Stage"] = Relationship(
        link_model=StageDependency,
        back_populates="children",
        sa_relationship_kwargs={
            "primaryjoin": "Stage.id==StageDependency.child_id",
            "secondaryjoin": "Stage.id==StageDependency.parent_id",
        },
    )
    children: List["Stage"] = Relationship(
        link_model=StageDependency,
        back_populates="parents",
        sa_relationship_kwargs={
            "primaryjoin": "Stage.id==StageDependency.parent_id",
            "secondaryjoin": "Stage.id==StageDependency.child_id",
        },
    )

    def attach_job(self, worker: Worker) -> Job:
        self.status = StageStatus.RUNNING
        job = Job(stage_id=self.id, worker_id=worker.id)
        self.jobs.append(job)
        return job



import pathlib
import shutil

import dvc.api
import pytest
import zntrack.examples

from paraffin.dvc import StageStatus, get_status


@pytest.fixture
def proj_single_node(proj_path):
    project = zntrack.Project()

    with project:
        stage1 = zntrack.examples.AddNumbers(a=1, b=2)

    project.build()

    return project


@pytest.fixture
def proj_nested_nodes(
    proj_path,
) -> tuple[
    zntrack.Project,
    list[zntrack.examples.AddNumbers | zntrack.examples.SumNodeAttributes],
]:
    project = zntrack.Project()

    with project:
        stage1 = zntrack.examples.AddNumbers(a=1, b=2)
        stage2 = zntrack.examples.AddNumbers(a=1, b=2)
        nested_1 = zntrack.examples.SumNodeAttributes(
            inputs=[stage1.c, stage2.c], shift=0
        )
        nested_2 = zntrack.examples.SumNodeAttributes(inputs=[nested_1.output], shift=0)
        nested_3 = zntrack.examples.SumNodeAttributes(inputs=[nested_2.output], shift=0)

    project.build()

    return project, [stage1, stage2, nested_1, nested_2, nested_3]


def test_stage_unfinished(proj_single_node):
    status = get_status()
    assert len(status) == 1
    stage = next(n for n in status if n.addressing == "AddNumbers")
    assert stage.status == StageStatus.PENDING

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    # # with repo.lock:
    status = repo.status()
    assert "AddNumbers" in status
    # stages = repo.stage.collect()
    # assert len(stages) == 1
    # for stage in stages:
    #     with stage.repo.lock:
    #         assert stage.changed() is True


def test_stage_finished(proj_single_node):
    proj_single_node.repro()

    status = get_status()
    assert len(status) == 1
    stage = next(n for n in status if n.addressing == "AddNumbers")
    assert stage.status == StageStatus.COMPLETED

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    # # with repo.lock:
    status = repo.status()
    assert status == {}
    # # stages = repo.stage.collect()
    # # assert len(stages) == 1
    # # for stage in stages:
    # #     with stage.repo.lock:
    # #         assert stage.changed() is False


@pytest.mark.parametrize("rmlock", [True, False])
def test_stage_cached(proj_single_node, rmlock):
    proj_single_node.repro()
    shutil.rmtree("nodes", ignore_errors=True)
    dvc_lock_path = pathlib.Path("dvc.lock")
    if rmlock:
        dvc_lock_path.unlink(missing_ok=True)
    # TODO: another check if the dvc.lock is removed!

    status = get_status()
    assert len(status) == 1
    stage = next(n for n in status if n.addressing == "AddNumbers")
    assert stage.status == StageStatus.COMPLETED

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    status = repo.status()
    # assert "AddNumbers" in status

    # stages = repo.stage.collect()
    # for stage in stages:
    #     with stage.repo.lock:
    #         try:
    #             stage.repo.stage_cache.restore(stage)
    #             # FYI, there is also stage.commit() which is like Stage.save(), but also saves file to the cache (i.e. commit).
    #             # Stage.dump() is what saves the stage to dvc.yaml and dvc.lock file.
    #             # (dump has update_pipeline=True|False and update_lock=True|False arguments to save to only one or to both of the files).
    #             stage.save()
    #             stage.dump()
    #         except RunCacheNotFoundError:
    #             raise ValueError("Unable to restore the stage")

    # status = repo.status()
    assert status == {}


def test_stage_cached_rm_cache(proj_single_node):
    proj_single_node.repro()
    shutil.rmtree("nodes", ignore_errors=True)
    pathlib.Path("dvc.lock").unlink(missing_ok=True)
    shutil.rmtree(".dvc/cache", ignore_errors=True)

    status = get_status()
    # assert status == {
    #     "AddNumbers": StageStatus.QUEUED,
    # }
    assert len(status) == 1
    stage = next(n for n in status if n.addressing == "AddNumbers")
    assert stage.status == StageStatus.PENDING

    # # TODO: another check if the dvc.lock is removed!

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    status = repo.status()
    assert "AddNumbers" in status

    # cache_available = None

    # stages = repo.stage.collect()
    # for stage in stages: # TODO: test later, that the order is correct!
    #     with stage.repo.lock:
    #         try:
    #             stage.repo.stage_cache.restore(stage)
    #             cache_available = True
    #         except RunCacheNotFoundError:
    #             cache_available = False
    #     # with stage.repo.lock:
    #     #     assert stage.changed() is False

    # assert cache_available is False
    # status = repo.status()
    # assert "AddNumbers" in status


def test_stage_nested_unfinished(proj_nested_nodes):
    status = get_status()
    proj, nodes = proj_nested_nodes
    for node in nodes:
        stage = next(s for s in status if s.addressing == node.name)
        assert stage.status == StageStatus.PENDING
    assert len(status) == 5


def test_stage_nested_finished(proj_nested_nodes):
    proj, nodes = proj_nested_nodes
    proj.repro()
    status = get_status()
    for node in nodes:
        stage = next(s for s in status if s.addressing == node.name)
        assert stage.status == StageStatus.COMPLETED, (
            f"Stage {node.name} is not completed"
        )
    assert len(status) == 5

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    # # with repo.lock:
    status = repo.status()
    assert status == {}


def test_stage_nested_cached(proj_nested_nodes):
    proj, nodes = proj_nested_nodes
    proj.repro()
    shutil.rmtree("nodes", ignore_errors=True)
    dvc_lock_path = pathlib.Path("dvc.lock")
    dvc_lock_path.unlink(missing_ok=True)

    status = get_status()
    for node in nodes:
        stage = next(s for s in status if s.addressing == node.name)
        assert stage.status == StageStatus.COMPLETED, (
            f"Stage {node.name} is not completed"
        )
    assert len(status) == 5

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    # # with repo.lock:
    status = repo.status()
    assert status == {}


def test_stage_nested_cached_rm_cache(proj_nested_nodes):
    proj, nodes = proj_nested_nodes
    proj.repro()
    shutil.rmtree("nodes", ignore_errors=True)
    dvc_lock_path = pathlib.Path("dvc.lock")
    dvc_lock_path.unlink(missing_ok=True)
    shutil.rmtree(".dvc/cache", ignore_errors=True)

    status = get_status()
    for node in nodes:
        stage = next(s for s in status if s.addressing == node.name)
        assert stage.status == StageStatus.PENDING, f"Stage {node.name} is not pending"
    assert len(status) == 5

    fs = dvc.api.DVCFileSystem()
    repo = fs.repo
    # # with repo.lock:
    status = repo.status()
    for node in nodes:
        assert node.name in status


def test_stage_nested_update_params_end(proj_nested_nodes):
    proj, nodes = proj_nested_nodes
    proj.repro()
    # update downstream node
    nodes[-1].shift = 1
    proj.build()

    status = get_status()
    for idx, node in enumerate(nodes):
        stage = next(s for s in status if s.addressing == node.name)
        if idx == len(nodes) - 1:
            assert stage.status == StageStatus.PENDING, (
                f"Stage {node.name} is not pending"
            )
        else:
            assert stage.status == StageStatus.COMPLETED, (
                f"Stage {node.name} is not completed"
            )


def test_stage_nested_update_params_between(proj_nested_nodes):
    proj, nodes = proj_nested_nodes
    proj.repro()
    # update downstream node
    nodes[-3].shift = 1
    proj.build()

    status = get_status()
    for idx, node in enumerate(nodes):
        stage = next(s for s in status if s.addressing == node.name)
        if idx >= len(nodes) - 2:
            # successors (two nodes) of the updated node
            assert stage.status == StageStatus.UNKNOWN, (
                f"Stage {node.name} is not unknown"
            )
        elif idx >= len(nodes) - 3:
            # this is the node that was updated
            assert stage.status == StageStatus.PENDING, (
                f"Stage {node.name} is not pending"
            )
        else:
            assert stage.status == StageStatus.COMPLETED, (
                f"Stage {node.name} is not completed"
            )



import dataclasses

import networkx as nx

from paraffin.db.app import StageStatus, query_existing_experiments
from paraffin.dvc import get_stage_from_graph


def prompt_transfer(stage_name: str, current_status: StageStatus) -> bool:
    print(f"Stage {stage_name} has status {current_status}")
    print(
        "Do you want to transfer the state of this stage to the new experiment? (Y/n)"
    )
    answer = input().lower().strip()
    return answer in ["", "y", "yes", "ja"]


def handle_existing_stages(graph, db):
    # Handle finished and unfinished stages
    for status in [StageStatus.FINISHED, StageStatus.UNFINISHED]:
        stages = query_existing_experiments(db_url=db, status=status, graph=graph)
        if not stages:
            continue

        print(f"Found {len(stages)} stages with {status} status from previous runs ")
        for stage in stages:
            should_transfer = prompt_transfer(stage.name, stage.status)
            new_status = stage.status if should_transfer else StageStatus.PENDING
            node = get_stage_from_graph(graph, stage.name)
            new_stage = dataclasses.replace(node, status=new_status)
            nx.relabel_nodes(graph, {node: new_stage}, copy=False)

    # Handle running stages
    running_stages = query_existing_experiments(
        db_url=db, status=StageStatus.RUNNING, graph=graph
    )
    if running_stages:
        print(
            f"Found {len(running_stages)} stages that are still running - "
            f"please stop the workers and run again."
        )
        return False

    return True



from paraffin.db.app import (
    # close_worker,
    # complete_job,
    # db_to_graph,
    # find_cached_job,
    # get_job,
    # get_job_dump,
    # get_jobs,
    # list_experiments,
    # list_workers,
    # register_worker,
    save_graph_to_db,
    #     update_job_status,
    #     update_worker,
)

__all__ = [
    "db_to_graph",
    "get_job_dump",
    "get_jobs",
    "list_experiments",
    "list_workers",
    "update_job_status",
    "close_worker",
    "complete_job",
    "find_cached_job",
    "get_job",
    "register_worker",
    "save_graph_to_db",
    "update_worker",
]
